#pragma semicolon 1

#include <sourcemod>
#include <regex>

#define PLUGIN_VERSION	"1.1"

new Handle:h_FilterType, i_filtertype,
	Handle:h_AllowLocal, bool:b_allowlocal,
	Handle:h_FilterPath, String:s_filterpath[PLATFORM_MAX_PATH];

public Plugin:myinfo = 
{
	name = "IPFilter",
	author = "FrozDark (HLModders LLC)",
	description = "Filters out black listed ip addresses",
	version = PLUGIN_VERSION,
	url = "http://www.hlmod.ru/"
};

public OnPluginStart()
{
	RegAdminCmd("sm_addipf", CommandBan, ADMFLAG_BAN, "sm_addipf <ip> <ip>");
	RegAdminCmd("sm_addipf_list", CommandList, ADMFLAG_BAN, "");
	h_FilterType = CreateConVar("sm_ipfilter_type", "2", "Filter type: 0-Disable|1-Allow only ipfilter list|2-Disallow only ipfilter list", 0, true, 0.0, true, 2.0);
	i_filtertype = GetConVarInt(h_FilterType);
	HookConVarChange(h_FilterType, OnConVarChange);
	
	h_AllowLocal = CreateConVar("sm_ipfilter_allow_local", "1", "Whether to allow local players to the server without checking", 0, true, 0.0, true, 1.0);
	b_allowlocal = GetConVarBool(h_AllowLocal);
	HookConVarChange(h_AllowLocal, OnConVarChange);
	
	h_FilterPath = CreateConVar("sm_ipfilter_path", "addons/sourcemod/data/ipfilter.txt", "Path to the ipfilter file");
	GetConVarString(h_FilterPath, s_filterpath, sizeof(s_filterpath));
	HookConVarChange(h_FilterPath, OnConVarChange);
	
	AutoExecConfig(true, "ipfilter");

	LoadTranslations("ipfilter.phrases");
}

public Action:CommandList(client, args)
{
	new Handle:file = OpenFile(s_filterpath, "r");
	if (file == INVALID_HANDLE)
	{
		LogError("Unable to open file '%s'", s_filterpath);
		return Plugin_Handled;
	}
	else
	{
		decl String:buffer[64];
		new x = 0;
		while (!IsEndOfFile(file) && ReadFileLine(file, buffer, sizeof(buffer)))
			PrintToConsole(client, "%d. %s", ++x, buffer);
	}
	CloseHandle(file);
	return Plugin_Handled;
}

public Action:CommandBan(client, args)
{
	decl String:buffer[100],
		 String:buffer2[100],
		 String:ipfilter[100];
	new Handle:ip = CompileRegex("\\d{1,3}.\\d{1,3}.\\d{1,3}.\\d{1,3}");
	
	GetCmdArg(1,buffer, sizeof(buffer));
	if (MatchRegex(ip, buffer) == 0)
	{
		ReplyToCommand(client, "[ipfilter] %t\n Usage: sm_addipf <ip> <ip>", "enter the correct ip address");
		return Plugin_Handled;
	}
		
	GetCmdArg(2,buffer2, sizeof(buffer2));
	if (MatchRegex(ip, buffer2) == 0)
	{
		ReplyToCommand(client, "[ipfilter] %t\n Usage: sm_addipf <ip> <ip>", "enter the correct ip address");
		return Plugin_Handled;
	}

	if (buffer[0] == '\0'||buffer2[0] == '\0')
	{
		ReplyToCommand(client, "[ipfilter] Usage: sm_addipf <ip> <ip>");
		return Plugin_Handled;
	}
	Format(ipfilter, sizeof(ipfilter), "%s - %s", buffer, buffer2);
	ReplaceString(ipfilter, sizeof(ipfilter), "\"", "", false);
	new Handle:hFile = OpenFile(s_filterpath, "at+"); 
	WriteFileLine(hFile, ipfilter);
	CloseHandle(hFile);
	PrintToChat(client, "[ipfilter] %t '%s'", "added ip subnetwork", ipfilter);
	LogAction(client,-1, "\"%L\" added ip subnetwork (%s)", client, ipfilter);
	
	return Plugin_Handled;
}

public OnConVarChange(Handle:convar, const String:oldValue[], const String:newValue[])
{
	if (convar == h_FilterType)
		i_filtertype = StringToInt(newValue);
	else if (convar == h_FilterPath)
		strcopy(s_filterpath, sizeof(s_filterpath), newValue);
	else if (convar == h_AllowLocal)
		b_allowlocal = bool:StringToInt(newValue);
}

public bool:OnClientConnect(client, String:rejectmsg[], maxlen)
{
	new bool:reject;
	switch (i_filtertype)
	{
		case 1 : reject = true;
		case 2 : reject = false;
		default : return true;
	}
	
	decl String:ip[17],
	String:buffer[2][64];
	GetClientIP(client, ip, sizeof(ip));
	
	if (b_allowlocal)
	{
		new long = IPToLong(ip);
		if (IsIPLocal(long))
			return true;
	}
	
	new client_ip = StringToInt(ip);
	
	new Handle:file = OpenFile(s_filterpath, "r");
	if (file == INVALID_HANDLE)
		return true;
	
	while (!IsEndOfFile(file))
	{
		ReadFileLine(file, buffer[0], sizeof(buffer[]));
		
		if (FindCharInString(buffer[0], '-') == -1) continue;
		
		ClearLine(buffer[0]);
	
		if (buffer[0][0] == '\0')
			continue;
		
		ExplodeString(buffer[0], "-", buffer, sizeof(buffer), sizeof(buffer[]));
		
		TrimString(buffer[0]);
		TrimString(buffer[1]);
		
		new ipfilter1 = StringToInt(buffer[0]);
		new ipfilter2 = StringToInt(buffer[1]);
		
		if (ipfilter2 == ipfilter1)
		{
			if (client_ip == ipfilter1)
			{
				reject = !reject;
				break;
			}
		}
		else if (ipfilter2 > ipfilter1)
		{
			if (ipfilter1 <= client_ip <= ipfilter2)
			{
				reject = !reject;
				break;
			}
		}
		else
		{
			if (ipfilter2 <= client_ip <= ipfilter1)
			{
				reject = !reject;
				break;
			}
		}
	}
	CloseHandle(file);
	
	if (reject)
	{
		Format(rejectmsg, maxlen, "%t", "KickByFilter");
		LogMessage("IP address '%s' filtered out from the server", ip);
	}
	return !reject;
}

ClearLine(String:g_szLine[])
{
	new pos;
	pos = StrContains((g_szLine), "//");
	if (pos != -1)
		g_szLine[pos] = '\0';

	pos = StrContains((g_szLine), "#");
	if (pos != -1)
		g_szLine[pos] = '\0';
		
	pos = StrContains((g_szLine), ";");
	if (pos != -1)
		g_szLine[pos] = '\0';
	
	TrimString(g_szLine);
}

/*
 * Converts a long IP to a dotted format String.
 * 
 * @param ip			IP Long
 * @param buffer		String Buffer (size = 16)
 * @param size			String Buffer size
 * @noreturn
 */
stock LongToIP(ip, String:buffer[], size)
{
	Format(
		buffer, size,
		"%d.%d.%d.%d",
			(ip >> 24)	& 0xFF,
			(ip >> 16)	& 0xFF,
			(ip >> 8 )	& 0xFF,
			ip        	& 0xFF
		);
}

/*
 * Converts a dotted format String IP to a long.
 * 
 * @param ip			IP String
 * @return				Long IP
 */
stock IPToLong(const String:ip[])
{
	decl String:pieces[4][4];

	if (ExplodeString(ip, ".", pieces, sizeof(pieces), sizeof(pieces[])) != 4)
		return 0;

	return (
		StringToInt(pieces[0]) << 24	|
		StringToInt(pieces[1]) << 16	|
		StringToInt(pieces[2]) << 8		|
		StringToInt(pieces[3])
	);
}

static localIPRanges[] = 
{
	10	<< 24,				// 10.
	127	<< 24 | 1		,	// 127.0.0.1
	127	<< 24 | 16	<< 16,	// 127.16.
	192	<< 24 | 168	<< 16,	// 192.168.
};

/*
 * Checks whether an IP is a private/internal IP
 * 
 * @param ip			IP Long
 * @return				True if the IP is local, false otherwise.
 */
stock bool:IsIPLocal(ip)
{
	new range, bits, move, bool:matches;

	for (new i=0; i < sizeof(localIPRanges); i++) 
	{

		range = localIPRanges[i];
		matches = true;

		for (new j=0; j < 4; j++) 
		{
			move = j * 8;
			bits = (range >> move) & 0xFF;

			if (bits && bits != ((ip >> move) & 0xFF)) 
				matches = false;
		}

		if (matches)
			return true;
	}

	return false;
}